#include <oxstd.oxh>
#include <oxdraw.oxh>
#include <oxprob.oxh>
#include <oxfloat.oxh>
#import <maximize>

////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
//
// Modeling Extreme Events: Time-varying Extreme Tail Shape.
// CODE FOR A SIMULATION EXAMPLE
//
// PACKAGE: OxMetrics 7
// DATE: 2020NOV11 
//
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

static decl s_mData, s_mXi, s_mDelta, s_mParm;	 

fKLfit(const dXi, const dDelta, const vXGrid) {
// compute the average log-likelihood via quasi MC integration, using the points in vXGrid
//
// dXi: scalar, tail index of the GPD
// dDelta: scalar, scale of the GPD
// vXGrid: grid of values to compute the log likelihood over
//
// Return: the average log likelihood over the grid
//
	return meanr( -log(dDelta) - (1 + 1/dXi) * log(1 + dXi * vXGrid / dDelta) );
}



fMakeXGrid(const aInvCdf, const dTailProbability, const iNrPoints) {
// compute grid of x values for quasi MC integration
//
// aInvCdf: pointer to a function, which is the inverse cdf; eats a matrix of q values, and
//          returns the corresponding q-quantiles of the distribution
// dTailprobability: tail area to construct the grid in, eg, 0.05 for the 5% *right* tail
// iNrPoints: number of points in the grid
//
// Return: 1 x iNrPoints vector with grid of x points
//
	decl vX = range(1,iNrPoints) / (1 + iNrPoints);
	decl dTau = aInvCdf(1 - dTailProbability);
	return { aInvCdf(1 - dTailProbability * vX) - dTau , dTau };
}



fFitDensity(const afQuantiles, const dTail, const dGridSize, const afStartingValues) {
// maximum likelihood estimation to fit the density of GPD distribution
//
// afQuantiles: pointer to a vector of quantiles
// dTail: tail area to construct the grid in, eg, 0.05 for the 5% *right* tail
// dGridSize: number of points in the grid
// afStartingValues: pointer to a function that generate starting values for MLE
//
// Return: 1 x 3 vector of estimated parameters
//
	decl vX, dTau;
	[vX, dTau] = fMakeXGrid(afQuantiles, dTail, dGridSize);
	decl dF=0, vP = afStartingValues(dTau);
	MaxControl(75,0,0);
	decl iReturn = MaxBFGS(
		[=](const vP, const adF, const avG, const amH) {
			adF[0] = double(fKLfit(exp(vP[0]), exp(vP[1]), vX));
			if (ismissing(adF[0])) return 0; else return 1;
		},
		&vP,
		&dF,
		0,
		1
	);
	if (iReturn != MAX_CONV) {
		println(MaxConvergenceMsg(iReturn));
		return { .NaN, .NaN, dTau};
	} else {		  
		return { exp(vP[0]) , exp(vP[1]) , dTau };
	}
}



fGasGPDFilter(const vData, const alphaD, const alphaX, const D0, const X0, const dLambda, const iReturnFilter=0) {
// GAS-GPD likelihood, and time varying parameters in the paper
//
// vData: 1 x n data input
// alphaD: parameter for Delta score update
// alphaX: parameter for Xi score update
// D0, X0: initial value of Delta and Xi, after transformation
// dLambda: smoothing parameter to control the decay of time varying parameters
// iReturnFilter: indicator whether to return the fitted time varying parameters
//
// Return:
//	if (iReturnFilter==0) return the likelihood
//  if (iReturnFilter==1) return the time varying parameters {Delta, Xi}
//
	decl dlik=0, vF, iT = columns(vData);
	decl dXi, dDelta;
	vF = log(X0 | D0);
	decl mA = diag(alphaX ~ alphaD);
	decl vXi, vDelta, vSmoothedScore=0;
	if (iReturnFilter == 1) vXi = vDelta = nans(vData);

	// run the filter
	for (decl i1=0;i1<iT;++i1) {
		// set parameters
		dXi = exp(vF[0]);
		dDelta = exp(vF[1]);
		if (iReturnFilter == 1) {vXi[i1] = dXi; vDelta[i1] = dDelta;}

		// is there a POT? if so, increase likelihood and update vF
		vSmoothedScore *= dLambda;
		if (vData[i1] > 0) {
		
			// compute likelihood contribution
			dlik += -log(dDelta) - (1 + 1/dXi)*log(1 + dXi * vData[i1] / dDelta);
			
			// update with GAS step
			vSmoothedScore += 
				(
					(1+dXi)*log(1 + dXi * vData[i1] / dDelta)/sqr(dXi)
					+
					(dDelta - (dXi + 3 + 1/dXi) * vData[i1])/(dDelta + dXi * vData[i1])
				)
				|
				(
					sqrt(1+2*dXi) * (vData[i1] - dDelta) / (dDelta + dXi * vData[i1])
				);
			if (ismissing(dlik) && (iReturnFilter!=1)) break;
			vF += mA * vSmoothedScore;
		} 
	}

	// return
	if (iReturnFilter==1) return {vDelta,vXi}; else return dlik;
	
}




fDynamicTau(const vData, const dTailProbability, const dSmoothingParameter=0.99) {
// time varying threshold 
//
// vData: 1 x n data input
// dTailProbability: tail probability to determine the tail region
// dSmoothingParameter: a_tau parameter in the dynamic tau process
//
// Return:
//	vTau: 1 x n vector of time varying tau
//
	decl vTau = nans(vData);
	vTau[0] = quantiler(vData[:minr(1000 ~ columns(vData)) - 1], 1 - dTailProbability);
//	vTau[0] = quantiler(vData[:25000], 1 - dTailProbability);
	for (decl i1=1;i1<columns(vData);++i1) {
		vTau[i1] = vTau[i1-1] + (1-dSmoothingParameter) * ((vData[i1-1] > vTau[i1-1]) -	dTailProbability);
	}
	return vTau;
}





fComputeTau(const vData, const dTailProbability, const iMethod=4,
	const vScaleOfData=1, const vTrueFilteredQuantiles=0, const dSmoothTau=0.99) { 
// different ways to generate thresholds 
//
// vData: 1 x n data input
// dTailProbability: tail probability to determine the tail region
// iMethod: the method to compute tau, default=4::time varying tau
// vScaleOfData: the scale of the data, result from the first-step GAS-T filter; default=1::no first-step filtering
// vTrueFilteredQuantiles: the true quantile as we know the DGP in simulations
// dSmoothTau: a_tau parameter in the dynamic tau process
//
// Return:
//	vTau: 1 x n vector of (time varying) tau
//
	decl vTau = nans(vData);
	if (iMethod == 1) {
		// true quantile
		vTau = vScaleOfData .* vTrueFilteredQuantiles;
	} else if (iMethod == 2) {
		// empirical quantile of true prefiltered data * true scale
		vTau = quantiler(vData ./ vScaleOfData, 1-dTailProbability) * vScaleOfData;
	} else if (iMethod == 3) {
		// iGAS time-varying quantile for true prefiltered data * true scale
		vTau = fDynamicTau(vData ./ vScaleOfData, dTailProbability, dSmoothTau) .* vScaleOfData;
//		Draw(0, vTau ./ vScaleOfData | vTrueFilteredQuantiles); ShowDrawWindow(); exit(0); 
	} else if (iMethod == 4) {
		// iGAS time-varying quantile for unfiltered data
		vTau = fDynamicTau(vData, dTailProbability, dSmoothTau);
	} else {
		// (static) empirical quantile of unfiltered data
		vTau = constant(quantiler(vData, 1-dTailProbability), vData);
	}
	return vTau;
}



////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
//
// ROUTINES FOR THE STUDENT T SETTING
//
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

fComputeStartingValuesStudentT(const dTau, const dSigma, const dNu) {
// compute the starting value for Xi , Delta in Student T simulations
 	return (1/dNu) | 1;	//	return (1 / dNu) | ((dNu * sqr(dSigma) + sqr(dTau)) / ((1+dNu) * dTau));  
}


fPseudoTrueXiDeltaStudentT(const vS, const vNu, const dTailProbability = 0.05, const dGridSize = 1e4,
	iSignOfLife = 0) {
// compute pseudo true values for vS , vNu
//
// vS: 1 x n vector of scale parameter of Student t
// vNu: 1 x n parameter vector df of Student t
// dTailProbability: tail area over which to minimize the KL divergence between
//         Student t and GPD; default = 5%
// dGridSize: grid size for quasi-MC integration; default 1e4; do not set too low
// iSignOfLife: if not 0, print progress
//
// return:
//  vDelta: 1 x n vector of pseudo true delta
//  vXi: 1 x n vector of pseudo true xi
//
	decl dXi, dDelta, dTau, dS, dNu;
	decl vXi = vS + .NaN;
	decl vDelta = vXi;

	for (decl i1=0;i1<columns(matrix(vS));++i1) {
	
		dS = vS[i1]; dNu = vNu[i1];
		// fit the density for this (dS2, dNu)
		[dXi, dDelta, dTau] = fFitDensity(
			[=](vX) {return dS * quant(vX, dNu);},
			dTailProbability,
			dGridSize,
			[=](dTau){return fComputeStartingValuesStudentT(dTau, dS, dNu);}
		);
		vXi[i1] = dXi; vDelta[i1] = dDelta;
		if (iSignOfLife != 0) if (fmod(i1,10) == 0) println(100*i1/columns(vS),"%%"," done ...");
	}
	return { vDelta, vXi };
}



fGenerateStudentT(const vS, const vNu) {
// simulate Student t distributed rv with time varying scale vS , and time varying df vNu
//
// vS: 1 x n vector of scale parameter of Student t
// vNu: 1 x n parameter vector df of Student t
//
// return: 1 x n vector of simulated data
//
	decl vData = vS;
	for (decl i1=0;i1<columns(vData);++i1) vData[i1] = rant(1,1,vNu[i1]);
	return vS .* vData;
}



fStudentTSimulationEstimation(const vS, const vNu, const dTail, const dGridSize,
	const iTauComputationMethod, const vTrueFilteredQuantiles, const dTauSmoothingParameter) {
// simulate Student t distributed rv and estimate the time varying tail model
//
// vS: 1 x n vector of scale parameter of Student t
// vNu: 1 x n parameter vector df of Student t
// dGridSize: grid size for quasi-MC integration; default 1e4; do not set too low
// iTauComputationMethod: the method to compute tau, default=4::time varying tau
// vTrueFilteredQuantiles: the true quantile as we know the DGP in simulations
// dTauSmoothingParameter: a_tau parameter in the dynamic tau process
//
// return:
//	vDeltaFit: 1 x n vector of estimated tail scale parameter Delta
//	vXiFit: 1 x n vector of estimated tail scale parameter Xi
//  vData: 1 x n vector of simulated data
//	vP: estimated parameters in the extreme tail shape model
//
	// generate data
	decl vData = fGenerateStudentT(vS, vNu);

	// generate thresholds
	decl dTau = fComputeTau(vData, dTail, iTauComputationMethod, vS, vTrueFilteredQuantiles, dTauSmoothingParameter);

	// compute starting values
	decl D0, X0;
	[D0, X0] = fPseudoTrueXiDeltaStudentT(vS[0], vNu[0], dTail, dGridSize);

	decl vP = <-2.3;-2.3;-8.4>, dF=0;
	MaxControl(250,100,1);
	decl iReturn =
		MaxBFGS(
			[=](const vP, const adF, const avG, const amH) {
				adF[0] = fGasGPDFilter(vData - dTau, exp(vP[1]), exp(vP[0]), D0, X0, 1/(1+exp(-vP[2])));
				if (ismissing(adF[0])) return 0; else return 1;
			},
			&vP, &dF, 0, 1
		);
	if (iReturn != MAX_CONV) println("SIMULATION: ",MaxConvergenceMsg(iReturn));
	decl vDeltaFit, vXiFit;
	[vDeltaFit, vXiFit] = fGasGPDFilter(vData - dTau, exp(vP[1]), exp(vP[0]), D0, X0, 1/(1+exp(-vP[2])), 1);
	println("Smoothing parameters <aX, aD, Lambda> = <",exp(vP[1]),",",exp(vP[0]),",",1/(1+exp(-vP[2])),">");
	return {vDeltaFit, vXiFit, vData, vP};
}






////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
//
// ROUTINES FOR THE GPD SETTING
//
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

quanGPD(const vX, const dDelta, const dXi) {
// inverse CDF of GPD with parameters (dDelta, dXi), input vX between 0 and 1
	return dDelta .* (((1 - vX).^(-dXi)) - 1) ./ dXi;
}


densGPD(const vX, const dDelta, const dXi) {
// density function of GPD with parameters (dDelta, dXi), input data vX
	return ((1 + dXi * vX / dDelta) .^ (-1 -1/dXi)) / dDelta;
}


fComputeStartingValuesGPD(const dTau, const dSigma2, const dNu) {
// compute the starting value for Xi , Delta in GPD simulations 
	return (1 / dNu) | (dSigma2 + ((1 / dNu) * dTau));
}


fPseudoTrueXiDeltaGPD(const vDeltaIn, const vXiIn, const dTailProbability = 0.05, const dGridSize = 1e4,
	iSignOfLife = 0) {
// compute pseudo true values for vS2 , vNu	but for the GPD case
//
// vDeltaIn: 1 x n vector of scale parameter of GPD
// vXiIn: 1 x n parameter vector, tail-index of GPD
// dTailProbability: tail area over which to minimize the KL divergence between
//         GPD and new GPD; default = 5%
// dGridSize: grid size for quasi-MC integration; default 1e4; do not set too low
// iSignOfLife: if not 0, print progress
//
// output:
// vDelta: 1 x n vector of pseudo true delta
// vXi: 1 x n vector of pseudo true xi
//
	decl dXi, dDelta, dTau, dXiIn, dDeltaIn;
	decl vXi = vXiIn + .NaN;
	decl vDelta = vXi;

	for (decl i1=0;i1<columns(matrix(vXiIn));++i1) {
	
		dXiIn = vXiIn[i1]; dDeltaIn = vDeltaIn[i1];
		// fit the density for this (dS2, dNu)
		[dXi, dDelta, dTau] = fFitDensity(
			[=](vX) {return quanGPD(vX, dDeltaIn, dXiIn);},
			dTailProbability,
			dGridSize,
			[=](dTau){return fComputeStartingValuesGPD(dTau, dDeltaIn, dXiIn);}
		);
		vXi[i1] = dXi; vDelta[i1] = dDelta;
		if (iSignOfLife != 0) if (fmod(i1,10) == 0) println(100*i1/columns(vXiIn),"%%"," done ...");
//		decl vX; [vX, dTau] = fMakeXGrid([=](vX) {return quanGPD(vX, dDeltaIn, dXiIn);}, dTailProbability, dGridSize); DrawXMatrix(0, densGPD(vX,dDelta,dXi) | (densGPD(vX+dTau,dDeltaIn,dXiIn) / dTailProbability),{"fit","true"},vX,0,0,2); ShowDrawWindow();
	}
	return { vDelta, vXi };
}


fGenerateGPD(vDelta, vXi) {
// generate 1 x n vector of GPD rv with parameters (vDelta, vXi)
	return quanGPD(ranu(1,columns(vDelta)), vDelta, vXi);
}


fGDPSimulationEstimation(const vDelta, const vXi, const dTail, const dGridSize,
	const iTauComputationMethod, const vTrueFilteredQuantiles, const dTauSmoothingParameter) {
// simulate GPD distributed rv and estimate the time varying tail model
//
// vDelta: 1 x n vector of tail scale parameter
// vXi: 1 x n vector of tail shape parameter
// dGridSize: grid size for quasi-MC integration; default 1e4; do not set too low
// iTauComputationMethod: the method to compute tau, default=4::time varying tau
// vTrueFilteredQuantiles: the true quantile as we know the DGP in simulations
// dTauSmoothingParameter: a_tau parameter in the dynamic tau process
//
// return:
//	vDeltaFit: 1 x n vector of estimated tail scale parameter Delta
//	vXiFit: 1 x n vector of estimated tail scale parameter Xi
//  vData: 1 x n vector of simulated data
//	vP: estimated parameters in the extreme tail shape model
//
	// generate data
	decl vData = fGenerateGPD(vDelta, vXi);

	// generate thresholds
	decl dTau = fComputeTau(vData, dTail, iTauComputationMethod, vDelta, vTrueFilteredQuantiles, dTauSmoothingParameter);

	// compute starting values
	decl D0, X0;
	[D0, X0] = fPseudoTrueXiDeltaGPD(vDelta[0], vXi[0], dTail, dGridSize);

	decl vP = <-2.3;-2.3;-3.4>, dF=0;
	MaxControl(250,100,1);
	decl iReturn =
		MaxBFGS(
			[=](const vP, const adF, const avG, const amH) {
				adF[0] = fGasGPDFilter(vData - dTau, exp(vP[1]), exp(vP[0]), D0, X0, 1/(1+exp(-vP[2])));
				if (ismissing(adF[0])) return 0; else return 1;
			},
			&vP, &dF, 0, 1
		);
	if (iReturn != MAX_CONV) println("SIMULATION: ",MaxConvergenceMsg(iReturn));
	decl vDeltaFit, vXiFit;
	[vDeltaFit, vXiFit] = fGasGPDFilter(vData - dTau, exp(vP[1]), exp(vP[0]), D0, X0, 1/(1+exp(-vP[2])), 1);
	println("Smoothing parameters <aX, aD, Lambda> = <",exp(vP[1]),",",exp(vP[0]),",",1/(1+exp(-vP[2])),">");
	return {vDeltaFit, vXiFit, vData, vP};
}



////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
//
// ROUTINES FOR SCENARIOS ON TIME VARYING PARAMETERS AND MAIN()
//
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

fXiDeltaForm(const iT, const iDGP) {
// set the DGP for the simulation
//
// iT: length of the sample
// iDGP: DGP number
//
// output:
// vDelta: a vector of true delta
// vXi: a vector of true xi
// vX: a vector of time
//
	decl vX, vXi, vDelta;
	
	vX = range(1,iT);

	switch_single (iDGP)
	{
		case 0:
			vXi = 0.5 * ones(vX);
			vDelta = 1 * ones(vX);
		case 1:
			vXi = 0.5 + 0.3 * sin(4 * M_PI * vX / iT);
			vDelta = 1 * ones(vX);
		case 2:
			vXi = 0.5 + 0.3 * sin(4 * M_PI * vX / iT);
			vDelta = 1 + 0.5 * sin(16 * M_PI * vX / iT);
		case 3:
			vXi = 0.5 + 0.3 * sin(4 * M_PI * vX / iT);
			vDelta = 1 + 0.5 * sin(4 * M_PI * vX / iT);
	}

	return {vXi, vDelta, vX};
}



SimulationMain(const GPDorSTUDENT, const iN, const iT, const iDGP, const PlotSignal) {
// main program of the simulation
//
// GPDorSTUDENT: indicator - simulate GPD dist. or T dist.
// iN: number of simulation
// iT: length of simulated sample
// iDGP: which DGP to simulate from
// PlotSignal: indicator - plot the outputs
//
// output:
//	A figure similar to Figure 2 in the manuscript. 
// 	Data, estimated parameters, estimated and true Xi/Delta saved as .xlsx files.
//
	// magic constants
	decl dTail = 0.05;
	decl dGridSize = 10000;
	decl dSmoothTau = 0.25;
	decl TauComputationMethod = 4;		// 0, or else: (static) empirical quantile of unfiltered data
										// 1: true quantile
										// 2: empirical quantile of true prefiltered data * true scale
										// 3: iGAS(dSmoothTau) time-varying quantile for true prefiltered data * true scale
										// 4: iGAS(dSmoothTau) time-varying quantile for unfiltered data

	// generate pattern for Xi and Delta
	decl vX, vXi, vDelta, vXCal, vXiCal, vDeltaCal;

	// compute pseudo true values on smaller grid size
	println("Start to compute pseudo true values of Delta and Xi ...");
	[vXi, vDelta, vX] = fXiDeltaForm(250, iDGP);
	vX *= iT / maxr(vX);
	decl vXiStar, vDeltaStar;
	if (GPDorSTUDENT == 0) [vDeltaStar, vXiStar] = fPseudoTrueXiDeltaGPD(vDelta, vXi, dTail, dGridSize, 0);
	else if (GPDorSTUDENT == 1)	[vDeltaStar, vXiStar] = fPseudoTrueXiDeltaStudentT(vDelta, 1 ./ vXi, dTail, dGridSize, 0);
	vXiCal = vXi; vDeltaCal = vDelta; vXCal = vX;
	println("Computation done! \n");
	
	[vXi, vDelta, vX] = fXiDeltaForm(iT, iDGP);
	s_mData = M_NAN*ones(iN, iT);
	s_mXi =	M_NAN*ones(iN, iT);
	s_mDelta = M_NAN*ones(iN, iT);
	s_mParm = M_NAN*ones(3, iN);
	
	// simulate data and filter back for one simulation
	decl vDeltaFit, vXiFit, vData, vP, i, vTrueQuantiles;
	for (i=0;i<iN;i++) 
	{
		ranseed(20201111+i);
		if (GPDorSTUDENT == 0) {
			vTrueQuantiles = quanGPD(1-dTail, 1, vXi);
			[vDeltaFit, vXiFit, vData, vP] = fGDPSimulationEstimation(vDelta, vXi, dTail, dGridSize, TauComputationMethod,
			vTrueQuantiles, dSmoothTau);
		} else if (GPDorSTUDENT == 1) {
			vTrueQuantiles = quant(1-dTail, 1 ./ vXi);
			[vDeltaFit, vXiFit, vData, vP] = fStudentTSimulationEstimation(vDelta, 1 ./ vXi, dTail, dGridSize, TauComputationMethod, vTrueQuantiles, dSmoothTau);
		}

		if (PlotSignal=="TRUE") {
			// plot result for filtered xi and delta
			DrawXMatrix(0, vXiFit, {"$\\xi_t^{Est}$"}, vX, 0, 0, 4);
			DrawXMatrix(1, vDeltaFit, {"$\delta_t^{Est}$"}, vX, 0, 0, 4);
		
			// plot pseudo true values
			DrawXMatrix(0, vXiStar, {"$\\xi_t^{True}$"}, vXCal, 0, 0, 2);
			DrawXMatrix(1, vDeltaStar, {"$\delta_t^{True}$"}, vXCal, 0, 0, 2);
			DrawAdjust(ADJ_AREAMATRIX,1,2);
			ShowDrawWindow();
		}

		s_mData[i][] = vData;
		s_mXi[i][] = vXiFit;
		s_mDelta[i][] = vDeltaFit;
		s_mParm[][i] = vP;
		
		savemat("Data_T"~sprint(iDGP)~"_Sim.xlsx", s_mData');
		savemat("Xi_T"~sprint(iDGP)~"_Est.xlsx", s_mXi');
		savemat("Delta_T"~sprint(iDGP)~"_Est.xlsx", s_mDelta');
	 	savemat("Parm_T"~sprint(iDGP)~"_Est.xlsx", s_mParm');
		savemat("Xi_T"~sprint(iDGP)~"_True.xlsx", vXiStar');
		savemat("Delta_T"~sprint(iDGP)~"_True.xlsx", vDeltaStar');
	}
	
}


main()
{
	decl GPDorSTUDENT, iN, iT, iDGP;

	GPDorSTUDENT = 1;	  // 0 for GPD, 1 for Student
	iN = 1; iT = 25000;		// iN: no. of simluation. iT: length of the simulated sample.
	iDGP = 2;				// DGP selection
	
	SimulationMain(GPDorSTUDENT, iN, iT, iDGP, "TRUE");		// Simulation routine

}